#include <gidx.h>
#include <gidx-wr.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>
#include <ctype.h>

#define MAX_LINE_BUF	8192
#define MAX_TOKS	1024

extern mesg_code_t _debug_level;
static const struct option l_ops[] = {
	{"schema", 1, 0, 's'},
	{"file", 1, 0, 'f'},
	{"output", 1, 0, 'o'},
	{"verbose", 0, 0, 'v'},
	{"terse", 0, 0, 't'},
	{"help", 0, 0, 'h'},
	{NULL, }
};
static const char * const s_ops = "s:f:o:vth";
static const char *cmd;
static const char *schema_fn;
static const char *input_fn;
static const char *output_fn;
static int verbose, terse;

_noreturn static void usage(int ecode)
{
	fprintf(stderr, "Usage: %s -s <schema> -f <infile> -o <outfile>\n",
		cmd);
	fprintf(stderr, "Generic gidx database creation utility.\n\n");
	fprintf(stderr, "    --schema,  -s     Schema file\n");
	fprintf(stderr, "    --file,    -f     Data file\n");
	fprintf(stderr, "    --output,  -o     Output file\n");
	fprintf(stderr, "    --verbose, -v     Verbose\n");
	fprintf(stderr, "    --terse,   -t     Terse\n");
	fprintf(stderr, "    --help,    -h     This message\n\n");

	exit(ecode);
}

static void parse_cmdline_opts(int argc, char **argv)
{
	int c;

	cmd = argv[0];

	while( (c = getopt_long(argc, argv, s_ops, l_ops, NULL)) != -1 ) {
		switch(c) {
		case 's':
			schema_fn = optarg;
			break;
		case 'f':
			input_fn = optarg;
			break;
		case 'o':
			output_fn = optarg;
			break;
		case 'v':
			verbose = 1;
			break;
		case 't':
			terse = 1;
			break;
		case 'h':
			usage(EXIT_SUCCESS);
		default:
			usage(EXIT_FAILURE);
		}
	}

	if ( NULL == schema_fn || NULL == input_fn || NULL == output_fn )
		usage(EXIT_FAILURE);
}

static int strtou32(const char *s, uint32_t *v)
{
	unsigned long int ret;
	char *end = NULL;

	ret=strtoul(s, &end, 0);

	if ( end == s )
		return -1;

	if ( ret == ULONG_MAX && errno == ERANGE )
		return -1;

	if ( ret > UINT_MAX )
		return -1;
	if ( ret & ~0xffffffff )
		return -1;

	*v = ret;

	return end - s;
}

/* Easy string tokeniser */
static int easy_explode(char *str, char split,
			char **toks, int max_toks)
{
	char *tmp;
	int tok;
	int state;

	for(tmp=str,state=tok=0; *tmp && tok <= max_toks; tmp++) {
		if ( state == 0 ) {
			if ( *tmp == split && (tok < max_toks)) {
				toks[tok++] = NULL;
			}else if ( !isspace(*tmp) ) {
				state = 1;
				toks[tok++] = tmp;
			}
		}else if ( state == 1 ) {
			if ( tok < max_toks ) {
				if ( *tmp == split || isspace(*tmp) ) {
					*tmp = '\0';
					state = 0;
				}
			}else if ( *tmp == '\n' )
				*tmp = '\0';
		}
	}

	return tok;
}

#define TOK_TYPE_NULL	0
#define TOK_TYPE_U32	1
#define TOK_TYPE_STR	2
struct tok {
	unsigned int t_type;
	union {
		uint32_t tu_uint;
		struct ro_vec tu_str;
	}t_u;
};

static int get_u32(void *rec, void *field, uint32_t *val)
{
	struct tok *tok = field;
	assert(tok->t_type == TOK_TYPE_U32);
	*val = tok->t_u.tu_uint;
	return 1;
}

static int get_str(void *rec, void *field, struct ro_vec *vec)
{
	struct tok *tok = field;
	assert(tok->t_type == TOK_TYPE_STR);
	*vec = tok->t_u.tu_str;
	return 1;
}

struct db {
	gidx_wr_t		db_wr;
	struct gidx_schema	*db_schema;
	unsigned int		db_num_fields;
};

#define SCHEMA_ALLOC_CHUNK   (1<<4)
#define SCHEMA_ALLOC_MASK    (SCHEMA_ALLOC_CHUNK-1)
static int schema_assure(struct gidx_schema **sch, unsigned int num)
{
	static void *new;

	if ( num & SCHEMA_ALLOC_MASK )
		return 1;

	new = realloc(*sch, sizeof(**sch) * (num + SCHEMA_ALLOC_CHUNK));
	if ( new == NULL )
		return 0;

	*sch = new;
	return 1;
}

static int parse_keytype(struct gidx_schema *s, char *tok)
{
	unsigned int i;
	static const char * const ktype[GIDX_NR_MODES] = {
		[GIDX_PRIMARY_KEY]	"PRIMARY",
		[GIDX_SECONDARY_KEY]	"SECONDARY",
	};

	for(i = 0; i < GIDX_NR_MODES; i++) {
		if ( ktype[i] && !strcasecmp(tok, ktype[i]) ) {
			s->s_mode = i;
			return 1;
		}
	}

	return 0;
}

static int parse_type(struct gidx_schema *s, char *tok)
{
	unsigned int i;
	static const struct {
		char * const name;
		union schema_get_val get_val;
	}ktype[GIDX_NR_TYPES] = {
		[GIDX_TYPE_U32]		{.name = "uint32_t",
						.get_val.su_get_uint = get_u32},
		[GIDX_TYPE_STRING]	{.name = "varchar",
						.get_val.su_get_str = get_str},
	};

	for(i = 0; i < GIDX_NR_TYPES; i++) {
		if ( ktype[i].name && !strcasecmp(tok, ktype[i].name) ) {
			s->s_type = i;
			s->s_get_val = ktype[i].get_val;
			return 1;
		}
	}

	return 0;
}

static int parse_schema_line(struct gidx_schema *s, char **tok, unsigned int nt)
{
	unsigned int idx = 0;

	if ( nt < 3 )
		return 0;
	if ( !parse_keytype(s, tok[0]) )
		return 0;
	if ( !parse_type(s, tok[1]) )
		return 0;
	s->s_name = strdup(tok[2]);
	if ( NULL == s->s_name )
		return 0;
	idx = 3;

	if ( idx < nt )
		WARN("unsupported options");
	return 1;
}

static struct gidx_schema *load_schema_file(const char *fn, unsigned int *num)
{
	char buf[MAX_LINE_BUF];
	char *tok[MAX_TOKS];
	struct gidx_schema *sch;
	unsigned int line, ns;
	FILE *f;

	mesg(M_INFO, "%s: loading schema from file: %s", cmd, fn);
	f = fopen(fn, "r");
	if ( NULL == f ) {
		ERR("fopen: %s", os_err2(NULL));
		return NULL;
	}

	for(sch = NULL, ns = 0, line = 1; fgets(buf, sizeof(buf), f); line++) {
		int num_toks;
		if ( buf[0] == '#' || buf[0] == '\r' || buf[0] == '\n' )
			continue;

		if ( NULL == strchr(buf, '\n') ) {
			ERR("line exceeded MAX_LINE_BUF (%u)", MAX_LINE_BUF);
			goto err;
		}

		num_toks = easy_explode(buf, 0, tok, MAX_TOKS);

		if ( !schema_assure(&sch, ns) ) {
			ERR("realloc: %s", os_err());
			goto err;
		}
		if ( !parse_schema_line(sch + ns, tok, num_toks) ) {
			ERR("%s:%u: parse error", fn, line);
			goto err;
		}
		ns++;
	}

	if ( !feof(f) )
		goto err;
	fclose(f);

	*num = ns;
	return sch;
err:
	fclose(f);
	free(sch);
	return NULL;
}

static struct db *db_new_from_schema(const char *fn)
{
	struct db *db;

	db = calloc(1, sizeof(*db));
	if ( NULL == db )
		goto out;

	db->db_schema = load_schema_file(fn, &db->db_num_fields);
	if ( NULL == db->db_schema )
		goto out_free_db;

	db->db_wr = gidx_wr_new(db->db_schema, db->db_num_fields);
	if ( NULL == db->db_wr )
		goto out_free_schema;

	return db;

out_free_schema:
	/* FIXME: strdup */
	free(db->db_schema);
out_free_db:
	free(db);
out:
	return NULL;
}

static void db_free(struct db *db)
{
	if ( db ) {
		/* FIXME: strdup */
		free(db->db_schema);
		free(db);
	}
}

static char *get_data_token(char *inp, struct tok *tok)
{
	while(*inp != '\0' && isspace(*inp))
		inp++;

	if ( *inp == '\0' )
		return inp;

	if ( !strncmp("NULL", inp, strlen("NULL")) ) {
		tok->t_type = TOK_TYPE_NULL;
		inp += strlen("NULL");
	}else if ( isdigit(*inp) ) {
		int dlen;

		tok->t_type = TOK_TYPE_U32;
		dlen = strtou32(inp, &tok->t_u.tu_uint);
		if ( dlen < 0 )
			return NULL;
		inp += dlen;
	}else if ( *inp == '\'' ) {
		char *tmp;

		for(tmp = inp + 1; *tmp != '\0' && *tmp != '\''; tmp++)
			/* nothing */;
		if ( *tmp == '\'' ) {
			*tmp = '\0';
			tok->t_type = TOK_TYPE_STR;
			tok->t_u.tu_str.v_ptr = (uint8_t *)inp + 1;
			tok->t_u.tu_str.v_len = tmp - (inp + 1);
			inp = tmp + 1;
		}else{
			return NULL;
		}
	}else{
		ERR("whoah (%c)", *inp);
		return NULL;
	}

	while(*inp != '\0' && isspace(*inp))
		inp++;

	return inp;
}

static int parse_and_insert(struct db *db, const char *fn,
				unsigned int line, char *buf)
{
	struct tok tok[db->db_num_fields];
	unsigned int nt, i;
	char *tmp;

	mesg(M_DEBUG, "data: %s:%u", fn, line);
	memset(&tok, 0, sizeof(tok));/* fuck off gcc */

	for(nt = 0, tmp = buf; (tmp = get_data_token(tmp, &tok[nt])); nt++) {
		if ( nt >= db->db_num_fields ) {
			WARN("too many fields");
			break;
		}

		switch(tok[nt].t_type) {
		case TOK_TYPE_NULL:
			mesg(M_DEBUG, "  %s -> NULL", 
				db->db_schema[nt].s_name);
			break;
		case TOK_TYPE_U32:
			mesg(M_DEBUG, "  %s -> %u", 
				db->db_schema[nt].s_name,
				tok[nt].t_u.tu_uint);
			break;
		case TOK_TYPE_STR:
			mesg(M_DEBUG, "  %s -> %.*s", 
				db->db_schema[nt].s_name,
				tok[nt].t_u.tu_str.v_len,
				(char *)tok[nt].t_u.tu_str.v_ptr);
			break;
		default:
			ERR("parse error, arg");
			return 0;
		}

		if ( *tmp == '\0' ) {
			if ( nt + 1 < db->db_num_fields )
				mesg(M_WARN, "%s:%i: not enough "
					"fields", fn, line);
			nt++;
			break;
		}
	}

	if ( NULL == tmp )
		return 0;

	gidx_wr_record_init(db->db_wr, buf);
	for(i = 0; i < nt; i++) {
		if ( tok[i].t_type == TOK_TYPE_NULL )
			continue;
		gidx_wr_record_set(db->db_wr, i, tok + i);
	}
	if ( !gidx_wr_record_commit(db->db_wr) )
		return 0;

	return 1;
}

static int db_insert_from_file(struct db *db, const char *fn)
{
	char buf[MAX_LINE_BUF];
	unsigned int line;
	FILE *f;

	mesg(M_WARN, "%s: loading schema from file: %s", cmd, fn);
	f = fopen(fn, "r");
	if ( NULL == f ) {
		ERR("fopen: %s", os_err2(NULL));
		return 0;
	}

	for(line = 1; fgets(buf, sizeof(buf), f); line++) {
		char *end;

		if ( buf[0] == '#' || buf[0] == '\r' || buf[0] == '\n' )
			continue;

		end = strchr(buf, '\n');
		if ( NULL == end ) {
			ERR("line exceeded MAX_LINE_BUF (%u)", MAX_LINE_BUF);
			goto err;
		}

		*end = '\0';
		if ( end - 1 >= buf && end[-1] == '\r' )
			*end = '\0';

		if ( !parse_and_insert(db, fn, line, buf) )
			goto err;
	}
	if ( !feof(f) )
		goto err;
	fclose(f);

	return 1;
err:
	fclose(f);
	return 0;
}

static int db_write_to_file(struct db *db, const char *outfile)
{
	int out_fd;

	mesg(M_WARN, "%s: output filename: %s", cmd, outfile);
	out_fd = open(outfile, O_WRONLY|O_CREAT|O_TRUNC, 0600);
	if ( out_fd < 0 ) {
		ERR("open: %s", os_err());
		return 0;
	}

	if ( !gidx_wr_write(db->db_wr, out_fd) )
		return 0;

	gidx_wr_free(db->db_wr);
	return 1;
}

int main(int argc, char **argv)
{
	struct db *db;

	parse_cmdline_opts(argc, argv);

	if ( verbose == terse ) {
		_debug_level = M_INFO;
	}else if ( verbose ) {
		_debug_level = M_DEBUG;
	}else if ( terse ) {
		_debug_level = M_WARN;
	}

	mesg(M_INFO, "gidx-create: constant database creation utility");
	mesg(M_INFO, "Copyright (c) 2010 Gianni Tedesco");
	mesg(M_INFO, "This program is free software; released under "
			"the terms of the GNU GPL v3 (see: COPYING)");
	
	db = db_new_from_schema(schema_fn);
	if ( NULL == db )
		return EXIT_FAILURE;
	
	if ( !db_insert_from_file(db, input_fn) )
		return EXIT_FAILURE;

	if ( !db_write_to_file(db, output_fn) )
		return EXIT_FAILURE;

	db_free(db);

	return EXIT_SUCCESS;
}
