/* TODO
 *  o check for oprhan chunks
 *  o check for overlapping chunks
*/
#include <gidx.h>
#include <gidx-format.h>
#include <unistd.h>

struct fsck {
	const struct gidx_hdr *hdr;
	const struct gidx_idx *idx;
	const struct gidx_chunk *chdr;
	const struct gidx_chunk *ptbl_chdr;
	const struct gidx_chunk *pkey_chdr;
	const uint8_t *strtab;
	const uint8_t *map;
	const uint8_t *map_end;
	size_t map_len;
};

static unsigned int strtab_sz(const struct gidx_hdr *hdr)
{
	return (hdr->h_strtab_sz_hi << 16) | hdr->h_strtab_sz_lo;
}

static unsigned int i_name(const struct gidx_idx *idx)
{
	return (idx->i_name_hi << 16) | idx->i_name_lo;
}

static int check_header(struct fsck *fsck)
{
	const struct gidx_hdr *hdr;
	gidx_off_t off;

	if ( fsck->map_len < sizeof(*fsck->hdr) )
		return 0;

	hdr = (struct gidx_hdr *)fsck->map;
	off = sizeof(*hdr);
	fsck->hdr = hdr;

	if ( hdr->h_magic != GIDX_MAGIC ) {
		ERR("bad magic: 0x%.8x", hdr->h_magic);
		return 0;
	}
	mesg(M_INFO, "hdr: magic = 0x%.8x", hdr->h_magic);

	if ( hdr->h_version != GIDX_VERSION ) {
		ERR("version mismatch: %u != %u", hdr->h_version, GIDX_VERSION);
		return 0;
	}

	mesg(M_INFO, "hdr: version %u.%u",
		(hdr->h_version >> 8), hdr->h_version & 0xff);

	if ( hdr->h_features & ~GIDX_ALL_FEATURES ) {
		ERR("Unspported features: 0x%.8x",
			hdr->h_features & ~GIDX_ALL_FEATURES);
		return 0;
	}

	fsck->idx = (struct gidx_idx *)(fsck->map + off);
	off += hdr->h_num_idx * sizeof(struct gidx_idx);
	if ( off > fsck->map_len ) {
		ERR("bad num_idx: %u", hdr->h_num_idx);
		fsck->idx = NULL;
		return 0;
	}
	mesg(M_INFO, "hdr: num_idx = %u", hdr->h_num_idx);

	if ( hdr->_pad0 )
		WARN("reserved field 0 set: 0x%.2x", hdr->_pad0);

	fsck->chdr = (struct gidx_chunk *)(fsck->map + off);
	off += hdr->h_num_chunks * sizeof(struct gidx_chunk);
	if ( off > fsck->map_len ) {
		ERR("bad num_chunks: %u", hdr->h_num_idx);
		fsck->chdr = NULL;
		return 0;
	}
	mesg(M_INFO, "hdr: num_chunks = %u", hdr->h_num_chunks);

	fsck->strtab = fsck->map + off;
	off += ALIGN_GIDX(strtab_sz(hdr));
	if ( off > fsck->map_len ) {
		ERR("bad strtab_sz: %u", strtab_sz(hdr));
		fsck->strtab = NULL;
		return 0;
	}
	mesg(M_INFO, "hdr: strtab = %u bytes", strtab_sz(hdr));

	if ( hdr->_pad1 )
		WARN("reserved field 1 set: 0x%.4x", hdr->_pad1);

	return 1;
}

static int check_chunk(struct fsck *fsck, unsigned int c, int sys)
{
	const struct gidx_chunk *ch;

	ch = fsck->chdr + c;

	if ( ch->c_data_ofs + ch->c_data_len > fsck->map_len ) {
		ERR("chunk[%u]: ofs %llu + %u bytes is beyond EOF",
			c, ch->c_data_ofs, ch->c_data_len);
		return 0;
	}

	if ( ch->_pad0 )
		WARN("chunk[%u]: reserved _pad0 field: 0x%.2x", c, ch->_pad0);
	if ( ch->_pad1 )
		WARN("chunk[%u]: reserved _pad1 field: 0x%.4x", c, ch->_pad1);

	/* TODO: set chunk bitmap */

	if ( !sys ) {
		mesg(M_INFO, "chunk[%u]: %u bytes @ offset %llu",
			c, ch->c_data_len, ch->c_data_ofs);
		mesg(M_INFO, "chunk[%u]: user_format: 0x%.2x",
			c, ch->c_format);
		hex_dump(fsck->map + ch->c_data_ofs, ch->c_data_len, 16);
		return 1;
	}

	switch(ch->c_format) {
	case GIDX_CHUNK_PKEY_TBL:
		fsck->ptbl_chdr = ch;
		mesg(M_INFO, "chunk[%u]: PKEY LOOKUP TABLE: "
			"%u bytes @ offset %llu",
			c, ch->c_data_len, ch->c_data_ofs);
		break;
	case GIDX_CHUNK_PKEY_BLOCK:
		fsck->pkey_chdr = ch;
		mesg(M_INFO, "chunk[%u]: PKEY BLOCK: "
			"%u bytes @ offset %llu",
			c, ch->c_data_len, ch->c_data_ofs);
		break;
	default:
		mesg(M_INFO, "chunk[%u]: UNKNOWN_SYS_CHUNK: 0x%.2x",
			c, ch->c_format);
		hex_dump(fsck->map + ch->c_data_ofs, ch->c_data_len, 16);
		break;
	}

	return 1;
}

static int check_pkey_tbl(const struct fsck *fsck,
				const struct gidx_chunk *tbl_hdr,
				const struct gidx_chunk *key_hdr)
{
	const struct gidx_pkey_tbl *tbl;
	const gidx_pkey_t *key;
	unsigned int num_tbl, num_key, i;

	tbl = (struct gidx_pkey_tbl *)(fsck->map + tbl_hdr->c_data_ofs);
	key = (gidx_pkey_t *)(fsck->map + key_hdr->c_data_ofs);

	num_tbl = tbl_hdr->c_data_len / sizeof(*tbl);
	num_key = key_hdr->c_data_len / sizeof(*key);

	mesg(M_INFO, "Primary key lookup table contents:");
	for(i = 0; i < num_tbl; i++) {
		unsigned int j;

		mesg(M_INFO, "  tbl[%u]: num_pkey=%u pkey_idx=%u",
			i, tbl[i].num_pkey, tbl[i].pkey_idx);
		for(j = 0; j < tbl[i].num_pkey; j++) {
			mesg(M_INFO, "   pkey[%u] = 0x%.8x",
				tbl[i].pkey_idx + j,
				key[tbl[i].pkey_idx + j]);
		}
	}

	return 1;
}

static int check_idx(struct fsck *fsck, unsigned int i)
{
 	const struct gidx_idx *idx;
	unsigned int n;

	idx = fsck->idx + i;

	if ( i_name(idx) + idx->i_nlen > strtab_sz(fsck->hdr) ) {
		ERR("index%u bad name %u bytes @ ofs %u",
			i, idx->i_nlen, i_name(idx));
		return 0;
	}

	if ( idx->i_type >= GIDX_NR_TYPES )
		WARN("index%u unknown type 0x%.2x", i, idx->i_type);
	if ( idx->i_mode >= GIDX_NR_MODES )
		WARN("index%u unknown mode 0x%.2x", i, idx->i_mode);
	if ( idx->_pad0 )
		WARN("index%u reserved field 1 set: 0x%.2x", i, idx->_pad0);

	mesg(M_INFO, "index[%u]: %.*s", i,
		idx->i_nlen,
		fsck->strtab + i_name(idx));
	
	if ( idx->i_num_chunks + idx->i_chunk_idx > fsck->hdr->h_num_chunks ) {
		ERR("index[%u]: too many chunks %u @ %u",
			i, idx->i_num_chunks, idx->i_chunk_idx);
		return 0;
	}

	mesg(M_INFO, "index[%u]:  o type %s", i, gidx_type_str(idx->i_type));
	mesg(M_INFO, "index[%u]:  o mode %s", i, gidx_mode_str(idx->i_mode));
	mesg(M_INFO, "index[%u]:  o format 0x%.2x", i, idx->i_format);
	mesg(M_INFO, "index[%u]:  o %u chunks @ index %u",
		i, idx->i_num_chunks, idx->i_chunk_idx);

	fsck->ptbl_chdr = NULL;
	fsck->pkey_chdr = NULL;
	for(n = 0; n < idx->i_num_chunks; n++) {
		int sys = (n < idx->i_sys_chunks);
		if ( !check_chunk(fsck, n + idx->i_chunk_idx, sys) )
			return EXIT_FAILURE;
	}

	if ( NULL == fsck->ptbl_chdr || NULL == fsck->pkey_chdr )
		return 1;
	if ( !check_pkey_tbl(fsck, fsck->ptbl_chdr, fsck->pkey_chdr) )
		return 0;

	return 1;
}

int main(int argc, char **argv)
{
	struct fsck fsck;
	unsigned int i;

	memset(&fsck, 0, sizeof(fsck));

	fsck.map = map_file(STDIN_FILENO, &fsck.map_len);
	if ( NULL == fsck.map ) {
		ERR("map_file: %s", os_err());
		return EXIT_FAILURE;
	}

	fsck.map_end = fsck.map + fsck.map_len;

	if ( !check_header(&fsck) )
		return EXIT_FAILURE;
	
	for(i = 0; i < fsck.hdr->h_num_idx; i++) {
		mesg(M_INFO, " ");
		if ( !check_idx(&fsck, i) )
			return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
