#include <gidx.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <gidx-rd.h>
#define _THIS_IS_LIBGIDX_RD 1
#include "gidx-rd-core.h"
#include <gidx-format.h>

static int mapfile(int fd, const uint8_t **begin, size_t *sz)
{
	struct stat st;

	if ( fstat(fd, &st) ) {
		ERR("fstat: %s", os_err());
		return 0;
	}

	*begin = mmap(NULL, st.st_size, PROT_READ,
			MAP_SHARED, fd, 0);
	if ( *begin == MAP_FAILED ) {
		ERR("mmap: %s", os_err());
		return 0;
	}

	*sz = st.st_size;

#ifdef MADV_RANDOM
	madvise((void *)*begin, st.st_size, MADV_RANDOM);
#endif
	return 1;
}

static unsigned int strtab_sz(const struct gidx_hdr *hdr)
{
	return (hdr->h_strtab_sz_hi << 16) | hdr->h_strtab_sz_lo;
}

static unsigned int i_name(const struct gidx_idx *idx)
{
	return (idx->i_name_hi << 16) | idx->i_name_lo;
}

static int check_header(struct _gidx_rd *rd)
{
	const struct gidx_hdr *hdr;
	gidx_off_t off;

	if ( rd->rd_maplen < sizeof(*hdr) )
		return 0;

	hdr = (struct gidx_hdr *)rd->rd_map;
	off = sizeof(*hdr);

	if ( hdr->h_magic != GIDX_MAGIC ) {
		ERR("bad magic: 0x%.8x", hdr->h_magic);
		return 0;
	}
	mesg(M_INFO, "hdr: magic = 0x%.8x", hdr->h_magic);

	if ( hdr->h_version != GIDX_VERSION ) {
		ERR("version mismatch: %u != %u", hdr->h_version, GIDX_VERSION);
		return 0;
	}

	mesg(M_INFO, "hdr: version %u.%u",
		(hdr->h_version >> 8), hdr->h_version & 0xff);

	if ( hdr->h_features & ~GIDX_ALL_FEATURES ) {
		ERR("Unspported features: 0x%.8x",
			hdr->h_features & ~GIDX_ALL_FEATURES);
		return 0;
	}

	rd->rd_idx = (struct gidx_idx *)(rd->rd_map + off);
	off += hdr->h_num_idx * sizeof(struct gidx_idx);
	if ( off > rd->rd_maplen ) {
		ERR("bad num_idx: %u", hdr->h_num_idx);
		rd->rd_idx = NULL;
		return 0;
	}
	mesg(M_INFO, "hdr: num_idx = %u", hdr->h_num_idx);

	if ( hdr->_pad0 )
		WARN("reserved field 0 set: 0x%.2x", hdr->_pad0);

	rd->rd_chunk = (struct gidx_chunk *)(rd->rd_map + off);
	off += hdr->h_num_chunks * sizeof(struct gidx_chunk);
	if ( off > rd->rd_maplen ) {
		ERR("bad num_chunks: %u", hdr->h_num_idx);
		rd->rd_chunk = NULL;
		return 0;
	}
	mesg(M_INFO, "hdr: num_chunks = %u", hdr->h_num_chunks);

	rd->rd_strtab = (rd->rd_map + off);
	off += ALIGN_GIDX(strtab_sz(hdr));
	if ( off > rd->rd_maplen ) {
		ERR("bad strtab_sz: %u", strtab_sz(hdr));
		rd->rd_strtab = NULL;
		return 0;
	}
	mesg(M_INFO, "hdr: strtab = %u bytes", strtab_sz(hdr));

	if ( hdr->_pad1 )
		WARN("reserved field 1 set: 0x%.4x", hdr->_pad1);

	rd->rd_num_idx = hdr->h_num_idx;
	rd->rd_num_chunk = hdr->h_num_chunks;
	return 1;
}

gidx_rd_t gidx_rd_open(int fd)
{
	struct _gidx_rd *rd;

	rd = calloc(1, sizeof(*rd));
	if ( NULL == rd )
		return NULL;;

	if ( !mapfile(fd, &rd->rd_map, &rd->rd_maplen) )
		goto out_free;

	if ( !check_header(rd) )
		goto out_unmap;

	return rd;
out_unmap:
	munmap((void *)rd->rd_map, rd->rd_maplen);
out_free:
	free(rd);
	return NULL;
}

gidx_idx_t gidx_rd_index(gidx_rd_t rd, const char *name)
{
	const struct gidx_idx *idx;
	struct ro_vec needle;
	unsigned int n;

	needle.v_ptr = (const uint8_t *)name;
	needle.v_len = strlen(name);

	for(n = rd->rd_num_idx, idx = rd->rd_idx; n; ) {
		struct ro_vec haystack;
	 	unsigned int i;
		int cmp;

		i = n / 2U;

		haystack.v_len = idx[i].i_nlen;
		haystack.v_ptr = rd->rd_strtab + i_name(idx + i);
		cmp = vcmp(&needle, &haystack);
		if ( cmp < 0 ) {
			n = i;
		}else if ( cmp > 0 ) {
			idx = idx + (i + 1U);
			n = n - (i + 1U);
		}else
			return idx + i;
	}

	return NULL;
}

void gidx_rd_fini(gidx_rd_t rd)
{
	if ( rd ) {
		munmap((void *)rd->rd_map, rd->rd_maplen);
		free(rd);
	}
}
