#include <gidx.h>
#include <gidx-wr.h>
#include <gang.h>
#include <gidx-format.h>
#define _THIS_IS_LIBGIDX_WR 1
#include "gidx-wr-core.h"
#define _THIS_IS_GIDX_STR_WR 1
#include "gidx-str-wr.h"

static struct str_val *get_node(struct str_wr_priv *strp, struct ro_vec *val)
{
	struct rb_node *n, *parent, **p;
	struct str_val *node;
	uint8_t *strbuf;

	DEBUG("searching for node val=%.*s", val->v_len, val->v_ptr);

	for(p = &strp->p_vals.rbt_root, n = parent = NULL; *p; ) {
		struct str_val *v;
		int ret;

		parent = *p;

		v = (struct str_val *)parent;
		ret = vcmp(val, &v->n_str);
		if ( ret < 0 ) {
			p = &(*p)->rb_child[CHILD_LEFT];
		}else if ( ret > 0 ) {
			p = &(*p)->rb_child[CHILD_RIGHT];
		}else{
			DEBUG("found node for val=%.*s, %u records",
				v->n_str.v_len,
				v->n_str.v_ptr,
				v->n_num_pkey);
			return v;
		}
	}

	node = hgang_alloc(strp->p_nodes);
	if ( NULL == node )
		return NULL;

	node->n_pkeys = hgang_new(sizeof(gidx_pkey_t), 128);
	if ( node->n_pkeys == NULL ) {
		ERR("hgang_new: %s", os_err());
		hgang_return(strp->p_nodes, node);
		return NULL;
	}

	strbuf = gang_alloc(strp->p_strtab, val->v_len);
	if ( NULL == strbuf ) {
		ERR("gang_alloc: %s", os_err());
		hgang_free(node->n_pkeys);
		hgang_return(strp->p_nodes, node);
		return NULL;
	}

	memcpy(strbuf, val->v_ptr, val->v_len);

	node->n_str.v_ptr = strbuf;
	node->n_str.v_len = val->v_len;
	node->n_num_pkey = 0;
	memset(&node->n_rbt, 0, sizeof(node->n_rbt));
	node->n_rbt.rb_parent = parent;
	*p = &node->n_rbt;
	rbtree_insert_rebalance(&strp->p_vals, &node->n_rbt);

	strp->p_num_vals++;

	DEBUG("inserted node for val=%.*s",
		node->n_str.v_len, node->n_str.v_ptr);
	return node;
}

static int node_pkey_id_insert(struct str_val *node, gidx_pkey_t pkey_id)
{
	gidx_pkey_t *ppkey;

	ppkey = hgang_alloc(node->n_pkeys);
	if ( NULL == ppkey ) {
		ERR("hgang_alloc: %s", os_err());
		return 0;
	}

	*ppkey = pkey_id;
	node->n_num_pkey++;

	DEBUG("append pkey_id=%u to node val=%.*s", pkey_id,
		node->n_str.v_len, node->n_str.v_ptr);
	return 1;
}

static int str_wr_insert(void *priv, gidx_pkey_t pkey_id,
			void *rec, void *field,
			union schema_get_val get_val)
{
	struct str_wr_priv *strp = priv;
	struct str_val *node;
	struct ro_vec vec;

	if ( !(*get_val.su_get_str)(rec, field, &vec) )
		return 0;

	node = get_node(strp, &vec);
	if ( node == NULL )
		return 0;

	if ( !node_pkey_id_insert(node, pkey_id) )
		return 0;

	DEBUG("priv=%p rec=%p val=%.*s", priv, rec, vec.v_len, vec.v_ptr);
	return 1;
}

static void *str_wr_init(gidx_mode_t mode, unsigned int options)
{
	struct str_wr_priv *strp;

	strp = calloc(1, sizeof(*strp));
	if ( NULL == strp )
		return NULL;

	strp->p_nodes = hgang_new(sizeof(struct str_val), 256);
	if ( NULL == strp->p_nodes ) {
		ERR("hgang_new: %s", os_err());
		free(strp);
		return NULL;
	}

	strp->p_strtab = gang_new(512, 0);
	if ( NULL == strp->p_strtab ) {
		ERR("gang_new: %s", os_err());
		hgang_free(strp->p_nodes);
		free(strp);
		return NULL;
	}

	strp->p_mode = mode;
	DEBUG("priv=%p", strp);
	return strp;
}

static void do_free(struct rb_node *n)
{
	struct str_val *node;

	if ( NULL == n )
		return;

	if ( n->rb_child[CHILD_LEFT] )
		do_free(n->rb_child[CHILD_LEFT]);

	node = (struct str_val *)n;
	DEBUG("node %.*s, %u pkeys",
		node->n_str.v_len, node->n_str.v_ptr, node->n_num_pkey);
	hgang_free(node->n_pkeys);

	if ( n->rb_child[CHILD_RIGHT] )
		do_free(n->rb_child[CHILD_RIGHT]);
}

static gidx_pkey_t str_wr_num_vals(void *priv)
{
	struct str_wr_priv *strp = priv;
	return strp->p_num_vals;
}

static void str_wr_fini(void *priv)
{
	struct str_wr_priv *strp = priv;

	DEBUG("priv=%p", priv);
	do_free(strp->p_vals.rbt_root);
	hgang_free(strp->p_nodes);
	gang_free(strp->p_strtab);
	if ( strp->p_prepped ) {
		switch(strp->p_format) {
		case STR_FORMAT_RADIX:
			_str_radix_free(strp);
			break;
		default:
			assert(0);
		}
	}
	free(priv);
}

static int do_write_lookup(struct rb_node *n, pkey_ctx_t pk)
{
	struct str_val *node;

	if ( NULL == n )
		return 1;

	if ( n->rb_child[CHILD_LEFT] )
		if ( !do_write_lookup(n->rb_child[CHILD_LEFT], pk) )
			return 0;

	node = (struct str_val *)n;
	if ( !pkey_tbl_entry(pk, node->n_num_pkey) )
		return 0;

	if ( n->rb_child[CHILD_RIGHT] )
		if ( !do_write_lookup(n->rb_child[CHILD_RIGHT], pk) )
			return 0;
	
	return 1;
}

static int str_wr_pkey_tbl(void *priv, pkey_ctx_t pk)
{
	struct str_wr_priv *strp = priv;
	return do_write_lookup(strp->p_vals.rbt_root, pk);
}

static int do_write_pkey(struct rb_node *n, pkey_ctx_t pk)
{
	struct str_val *node;

	if ( NULL == n )
		return 1;

	if ( n->rb_child[CHILD_LEFT] )
		if ( !do_write_pkey(n->rb_child[CHILD_LEFT], pk) )
			return 0;

	node = (struct str_val *)n;
	pkey_keyblock(pk, node->n_pkeys);

	if ( n->rb_child[CHILD_RIGHT] )
		if ( !do_write_pkey(n->rb_child[CHILD_RIGHT], pk) )
			return 0;

	return 1;
}

static int str_wr_pkey_blks(void *priv, pkey_ctx_t pk)
{
	struct str_wr_priv *strp = priv;
	return do_write_pkey(strp->p_vals.rbt_root, pk);
}

static uint16_t str_wr_num_chunks(void *priv)
{
	struct str_wr_priv *strp = priv;
	switch(strp->p_format) {
	case STR_FORMAT_RADIX:
		return _str_radix_num_chunks(strp);
	default:
		ERR("wtf");
		return 0;
	}
	return 0;
}

static off_t str_wr_chunk_size(void *priv, unsigned int chunk_id)
{
	struct str_wr_priv *strp = priv;
	switch(strp->p_format) {
	case STR_FORMAT_RADIX:
		return _str_radix_chunk_size(strp, chunk_id);
	default:
		ERR("wtf");
		return 0;
	}
	return 0;
}

static int str_wr_chunk(void *priv, unsigned int chunk_id, fobuf_t buf)
{
	struct str_wr_priv *strp = priv;
	switch(strp->p_format) {
	case STR_FORMAT_RADIX:
		return _str_radix_chunk(strp, chunk_id, buf);
	default:
		ERR("wtf");
		return 0;
	}
	return 1;
}

static int str_wr_prep(void *priv, uint8_t *format)
{
	struct str_wr_priv *strp = priv;

	strp->p_format = STR_FORMAT_RADIX;
	switch(strp->p_format) {
	case STR_FORMAT_RADIX:
		if ( !_str_radix_prep(strp) )
			return 0;
		break;
	default:
		ERR("wtf");
		return 0;
	}

	(*format) = strp->p_format;
	return 1;
}

struct _gidx_type_wr _gidx_str_type = {
	.t_name = "varchar",
	.t_wr_init = str_wr_init,
	.t_wr_fini = str_wr_fini,
	.t_wr_insert = str_wr_insert,
	.t_wr_prep = str_wr_prep,
	.t_wr_num_chunks = str_wr_num_chunks,
	.t_wr_chunk_size = str_wr_chunk_size,
	.t_wr_chunk = str_wr_chunk,
	.t_wr_num_vals = str_wr_num_vals,
	.t_wr_pkey_tbl = str_wr_pkey_tbl,
	.t_wr_pkey_blks = str_wr_pkey_blks,
};
