#include <gidx.h>
#include <stdio.h>
#include <sys/types.h> /* for lseek */
#include <unistd.h> /* for lseek */
#include <gidx-wr.h>
#include <gidx-format.h>
#define _THIS_IS_LIBGIDX_WR 1
#include "gidx-wr-core.h"
#define _THIS_IS_GIDX_WR_C 1
#include "gidx-pkey.h"

struct wr_field {
	gidx_pkey_t		f_num_recs;
	union schema_get_val 	f_get_val;
	struct _gidx_type_wr 	*f_type;
	void			*f_priv;
	const char 		*f_name;
	struct wr_chunk		*f_chunk;
	unsigned int		f_num_chunks;
	uint16_t		f_sys_chunks;
	gidx_type_t		f_type_id;
	gidx_mode_t		f_mode;
	uint8_t			f_format;
	void			*f_ctx; /* like f_priv but for sys chunks */
};

struct wr_chunk {
	gidx_off_t		c_ofs;
	struct wr_field 	*c_field;
	int			c_blkid;
	uint8_t			c_format;
	uint32_t 		c_len;
};

struct wr_rec {
	void 			*r_record;
	void			*r_field[0];
};

struct _gidx_wr {
	gidx_pkey_t		wr_num_records;
	struct wr_rec		*wr_rec;
	struct wr_field 	*wr_field;
	struct wr_chunk		*wr_chunk;
	uint8_t			wr_num_fields;
	uint32_t		wr_tot_chunks;
};

_public gidx_wr_t gidx_wr_new(const struct gidx_schema *schema,
				unsigned int num)
{
	struct _gidx_wr *wr;
	unsigned int i;

	if ( num > GIDX_MAX_INDEX ) {
		ERR("num_fields (%u) greater than maximum permitted (%u)",
			num, GIDX_MAX_INDEX);
		return 0;
	}

	wr = calloc(1, sizeof(*wr));
	if ( NULL == wr ) {
		ERR("calloc: %s", os_err());
		return NULL;
	}

	wr->wr_field = calloc(num, sizeof(*wr->wr_field));
	if ( NULL == wr->wr_field ) {
		ERR("calloc: %s", os_err());
		goto out;
	}

	for(i = 0; i < num; i++) {
		/* Field name */
		if ( NULL == schema[i].s_name ||
			strlen(schema[i].s_name) > GIDX_MAX_INDEX_NLEN ) {
			ERR("%s: bad field name", schema[i].s_name);
			goto out_free_field;
		}
		wr->wr_field[i].f_name = schema[i].s_name;

		/* get_val */
		if ( NULL == schema[i].s_get_val.su_get_str ) {
			ERR("%s: get_val unset", schema[i].s_name);
			goto out_free_field;
		}
		wr->wr_field[i].f_get_val = schema[i].s_get_val;

		/* field type */
		if ( schema->s_type >= GIDX_NR_TYPES ) {
			ERR("%s: bad type (%u)", schema[i].s_name,
						schema[i].s_type);
			goto out_free_field;
		}
		wr->wr_field[i].f_type = _gidx_types[schema[i].s_type];
		wr->wr_field[i].f_type_id = schema[i].s_type;
		wr->wr_field[i].f_mode = schema[i].s_mode;
		wr->wr_num_fields++;
	}

	/* Initialise field */
	for(i = 0; i < num; i++) {
		struct _gidx_type_wr *type;
		gidx_mode_t mode;

		type = wr->wr_field[i].f_type;
		mode = wr->wr_field[i].f_mode;

		wr->wr_field[i].f_priv =(*type->t_wr_init)(mode,
						schema[i].s_options);
		if ( NULL == wr->wr_field[i].f_priv ) {
			num = i;
			goto out_free_priv;
		}

		DEBUG("field[%u]: %s %s %s", i,
			gidx_mode_str(mode),
			type->t_name,
			wr->wr_field[i].f_name);
	}

	wr->wr_rec = malloc(wr->wr_num_fields * sizeof(*wr->wr_rec->r_field) +
				sizeof(*wr->wr_rec));
	if ( NULL == wr->wr_rec )
		goto out_free_rec;
	
	wr->wr_rec->r_record = NULL;

	return wr;

out_free_rec:
	free(wr->wr_rec);
out_free_priv:
	for(i = 0; i < num; i++) {
		struct _gidx_type_wr *type;
		type = wr->wr_field[i].f_type;
		type->t_wr_fini(wr->wr_field[i].f_priv);
		DEBUG("free field %u", i);
	}
out_free_field:
	free(wr->wr_field);
out:
	free(wr);
	return NULL;
}

void gidx_wr_record_init(gidx_wr_t wr, void *record)
{
	unsigned int i;
	assert(NULL == wr->wr_rec->r_record);
	wr->wr_rec->r_record = record;
	for(i = 0; i < wr->wr_num_fields; i++)
		wr->wr_rec->r_field[i] = NULL;
}

void gidx_wr_record_set(gidx_wr_t wr, unsigned int field, void *priv)
{
	assert(priv);
	assert(wr->wr_rec->r_record);
	assert(field < wr->wr_num_fields);
	wr->wr_rec->r_field[field] = priv;
}

int gidx_wr_record_commit(gidx_wr_t wr)
{
	unsigned int i;
	assert(wr->wr_rec->r_record);
	DEBUG("COMMIT");
	for(i = 0; i < wr->wr_num_fields; i++) {
		if ( NULL == wr->wr_rec->r_field[i] )
			continue;
			
		DEBUG("field: -- %s", wr->wr_field[i].f_name);
		if ( !(*wr->wr_field[i].f_type->t_wr_insert)(
					wr->wr_field[i].f_priv,
					wr->wr_num_records,
					wr->wr_rec->r_record,
					wr->wr_rec->r_field[i],
					wr->wr_field[i].f_get_val) )
			return 0;
		wr->wr_field[i].f_num_recs++;
	}
	wr->wr_num_records++;
	wr->wr_rec->r_record = NULL;
	return 1;
}

void gidx_wr_record_abort(gidx_wr_t wr)
{
	assert(wr->wr_rec->r_record);
	wr->wr_rec->r_record = NULL;
}

static int write_header(fobuf_t buf, uint8_t num_idx,
			uint32_t total_chunks,
			size_t strtab_sz)
{
	struct gidx_hdr hdr;

	assert(strtab_sz < (1<<24));

	hdr.h_magic = GIDX_MAGIC;
	hdr.h_version = GIDX_VERSION;
	hdr.h_features = GIDX_ALL_FEATURES;
	hdr.h_num_idx = num_idx;
	hdr._pad0 = 0;
	hdr.h_num_chunks = total_chunks;
	hdr.h_strtab_sz_hi = (strtab_sz) >> 16;
	hdr.h_strtab_sz_lo = strtab_sz & 0xff;
	hdr._pad1 = 0;

	return fobuf_write(buf, &hdr, sizeof(hdr));
}

static int write_index(fobuf_t buf, struct wr_field *f,
			unsigned int *c_idx, unsigned int *n_ofs)
{
	struct gidx_idx idx;

	assert((*n_ofs) + strlen(f->f_name) < (1<<24));

	idx.i_chunk_idx = *c_idx;
	idx.i_num_chunks = f->f_num_chunks;
	idx.i_nlen = strlen(f->f_name);
	idx.i_type = f->f_type_id;
	idx.i_mode = f->f_mode;
	idx.i_format = f->f_format;
	idx._pad0 = 0;
	idx.i_name_hi = (*n_ofs) >> 16;
	idx.i_name_lo = (*n_ofs) & 0xff;
	idx.i_sys_chunks = f->f_sys_chunks;

	(*n_ofs) += idx.i_nlen;
	(*c_idx) += f->f_num_chunks;

	return fobuf_write(buf, &idx, sizeof(idx));
}

#define SKEY_NUM_SYS_CHUNKS 2
static int size_up_chunks(struct _gidx_wr *wr)
{
	unsigned int i, c, num_chunks;

	for(i = num_chunks = 0; i < wr->wr_num_fields; i++) {
		struct wr_field *f;

		f = wr->wr_field + i;
		f->f_num_chunks = (*f->f_type->t_wr_num_chunks)(f->f_priv);
		f->f_sys_chunks = SKEY_NUM_SYS_CHUNKS;
		DEBUG("%s: %s: %u system chunks + %u chunks requested",
			f->f_type->t_name, f->f_name,
			f->f_sys_chunks, f->f_num_chunks);
		f->f_num_chunks += f->f_sys_chunks;
		num_chunks += f->f_num_chunks;
	}

	wr->wr_chunk = calloc(num_chunks, sizeof(*wr->wr_chunk));
	if ( NULL == wr->wr_chunk ) {
		ERR("calloc: %s", os_err());
		return 0;
	}

	wr->wr_tot_chunks = num_chunks;

	for(i = c = 0; i < wr->wr_num_fields; i++) {
		struct wr_field *f;
		struct wr_chunk *ch;
		unsigned int j, num;

		f = wr->wr_field + i;
		f->f_chunk = wr->wr_chunk + c;

		ch = wr->wr_chunk + c + 0;
		ch->c_field = f;
		ch->c_blkid = -1;
		ch->c_format = GIDX_CHUNK_PKEY_TBL;
		ch->c_len = sizeof(struct gidx_pkey_tbl) *
				(*f->f_type->t_wr_num_vals)(f->f_priv);

		ch = wr->wr_chunk + c + 1;
		ch->c_field = f;
		ch->c_blkid = -1;
		ch->c_format = GIDX_CHUNK_PKEY_BLOCK;
		ch->c_len = sizeof(gidx_pkey_t) * f->f_num_recs;

		c += SKEY_NUM_SYS_CHUNKS;

		num = (*f->f_type->t_wr_num_chunks)(f->f_priv);
		for(j = 0; j < num; j++, c++) {
			ch = wr->wr_chunk + c;
			ch->c_field = f;
			ch->c_blkid = j;
			if ( f->f_type->t_wr_chunk_format )
				ch->c_format = (*f->f_type->t_wr_chunk_format)
							(f->f_priv, j);
			else
				ch->c_format = GIDX_CHUNK_USER;
			ch->c_len = (*f->f_type->t_wr_chunk_size)
						(f->f_priv, ch->c_blkid);
		}
	}

	assert(c == wr->wr_tot_chunks);
	return 1;
}

static int c_cmp(const void *A, const void *B)
{
	struct wr_chunk * const * const a1 = A, * const * const b1 = B;
	const struct wr_chunk *a = (*a1), *b = (*b1);

	if ( a->c_blkid < 0 && b->c_blkid < 0 ) {
		/* compare 2 system blocks */
		if ( a->c_format == b->c_format ) {
			assert(a->c_field != b->c_field);
			return a->c_field - b->c_field;
		}
		return a->c_format - b->c_format;
	}
	if ( a->c_blkid >= 0 && b->c_blkid >= 0 ) {
		if ( a->c_field == b->c_field ) {
			assert(a->c_blkid != b->c_blkid);
			return a->c_blkid - b->c_blkid;
		}
		return a->c_field - b->c_field;
	}
	if ( b->c_blkid < 0 )
		return -1;
	if ( a->c_blkid > 0 )
		return 1;
	assert(1);
	return 0;
}

static gidx_off_t layout_chunks(struct wr_chunk **c, unsigned int num_chunks,
				gidx_off_t chunk_base)
{
	gidx_off_t off = chunk_base;
	gidx_off_t data_sz = 0;
	unsigned int i;

	sort(c, num_chunks, sizeof(*c), c_cmp);

	for(i = 0; i < num_chunks; i++) {
		const struct wr_field *f;
		struct wr_chunk *ch;

		ch = c[i];
		f = ch->c_field;
		ch->c_ofs = off;
		INFO("%u: %s: chunk[%u/%i] %u bytes @ %llu", i,
			f->f_name, ch - f->f_chunk, ch->c_blkid,
			ch->c_len, ch->c_ofs);
		off += ALIGN_GIDX(ch->c_len);
		data_sz += off;
	}

	return data_sz;
}

static int write_chdr(fobuf_t buf, struct wr_chunk *chunk)
{
	struct gidx_chunk ch;

	ch.c_data_ofs = chunk->c_ofs;
	ch.c_data_len = chunk->c_len;
	ch.c_format = chunk->c_format;
	ch._pad0 = 0;
	ch._pad1 = 0;

	return fobuf_write(buf, &ch, sizeof(ch));
}

#define CHECK_WRITES 1
#if CHECK_WRITES
struct wrcheck {
	fobuf_t buf;
	off_t begin;
	off_t end;
};
static void wrcheck_init(struct wrcheck *wr, fobuf_t buf)
{
	if ( fobuf_flush(buf) ) {
		wr->buf = buf;
		wr->begin = lseek(fobuf_fd(buf), 0, SEEK_CUR);
	}else{
		wr->buf = NULL;
	}
}
_check_result static int wrcheck_check(struct wrcheck *wr, off_t c_len)
{
	if ( NULL == wr->buf )
		return 0;

	if ( fobuf_flush(wr->buf) ) {
		wr->end = lseek(fobuf_fd(wr->buf), 0, SEEK_CUR);
		if ( wr->end - wr->begin != c_len ) {
			ERR("truncated write: %llu != %llu",
				wr->end - wr->begin, c_len);
			return 0;
		}
	}

	return 1;
}
#else
struct wrcheck{};
static void wrcheck_init(struct wrcheck *wr, fobuf_t buf) {}
static int wrcheck_check(struct wrcheck *wr, gidx_off_t c_len) { return 1; }
#endif

static const char alignbuf[GIDX_ALIGN - 1];

int gidx_wr_align(fobuf_t buf, size_t obj_sz)
{
	return fobuf_write(buf, alignbuf, GIDX_ALIGN_BYTES(obj_sz));
}

static int write_pk_tbl(fobuf_t buf, struct _gidx_wr *wr,
				const struct wr_chunk *ctbl)
{
	struct _pkey_ctx *ctx;
	struct wrcheck wrc;
	struct wr_field *f;
	size_t align;

	assert(ctbl->c_format == GIDX_CHUNK_PKEY_TBL);
	f = ctbl->c_field;
	assert(NULL == f->f_ctx);
	align = GIDX_ALIGN_BYTES(ctbl->c_len);

	ctx = malloc(sizeof(*ctx));
	if ( NULL == ctx )
		return 0;

	f->f_ctx = ctx;

	if ( !pkey_init(ctx, wr, buf) )
		return 0;
	wrcheck_init(&wrc, buf);
	if ( !(*f->f_type->t_wr_pkey_tbl)(f->f_priv, ctx) )
		return 0;
	if ( !wrcheck_check(&wrc, ctbl->c_len) )
		return 0;

	return fobuf_write(buf, alignbuf, align);
}

static int write_pk_block(fobuf_t buf, struct _gidx_wr *wr,
				const struct wr_chunk *ckey)
{
	struct _pkey_ctx *ctx;
	struct wrcheck wrc;
	struct wr_field *f;
	size_t align;

	assert(ckey->c_format == GIDX_CHUNK_PKEY_BLOCK);
	f = ckey->c_field;
	ctx = f->f_ctx;
	assert(ctx);
	align = GIDX_ALIGN_BYTES(ckey->c_len);

	if ( !pkey_init2(ctx) )
		goto err;
	wrcheck_init(&wrc, buf);
	if ( !(*f->f_type->t_wr_pkey_blks)(f->f_priv, ctx) )
		goto err;
	if ( !wrcheck_check(&wrc, ckey->c_len) )
		goto err;
	if ( !pkey_fini(ctx) )
		goto err;

	free(ctx);
	f->f_ctx = NULL;
	return fobuf_write(buf, alignbuf, align);
err:
	free(ctx);
	f->f_ctx = NULL;
	return 0;
}

static int write_user_chunk(fobuf_t buf, struct wr_chunk *ch)
{
	size_t align;
	struct wr_field *f;
	struct wrcheck wrc;

	wrcheck_init(&wrc, buf);

	f = ch->c_field;

	align = GIDX_ALIGN_BYTES(ch->c_len);
	if ( !(*f->f_type->t_wr_chunk)(f->f_priv, ch->c_blkid, buf) )
		return 0;

	if ( !wrcheck_check(&wrc, ch->c_len) )
		return 0;

	DEBUG("%s: chunk%u: wrote %u bytes, %u bytes alignment",
		f->f_name, ch->c_blkid, ch->c_len, align);
	return fobuf_write(buf, alignbuf, align);
}

static size_t do_prep(struct _gidx_wr *wr)
{
	unsigned int i;
	size_t strtab_sz;

	for(strtab_sz = i = 0; i < wr->wr_num_fields; i++) {
		struct wr_field *f;

		f = wr->wr_field + i;

		if ( f->f_type->t_wr_prep &&
			!(*f->f_type->t_wr_prep)(f->f_priv, &f->f_format) ) {
			ERR("%s: prep_failed on field", f->f_name);
			return 0;
		}

		DEBUG("%s: storage format 0x%.2x", f->f_name, f->f_format);
		strtab_sz += strlen(f->f_name);
	}

	return strtab_sz;
}

static int f_cmp(const void *A, const void *B)
{
	struct wr_field * const * const a = A, * const * const b = B;
	return strcmp((*a)->f_name, (*b)->f_name);
}

_public int gidx_wr_write(gidx_wr_t wr, int fd)
{
	size_t strtab_sz, idxblk_sz, chblk_sz;
	unsigned int i, j, c_idx, s_ofs;
	struct wr_field **flayout = NULL;
	struct wr_chunk **clayout = NULL;
	gidx_off_t first_chunk, data_sz;
	size_t align;
	fobuf_t buf;

	idxblk_sz = wr->wr_num_fields * sizeof(struct gidx_idx);
	strtab_sz = do_prep(wr);
	if ( 0 == strtab_sz )
		return 0;

	if ( !size_up_chunks(wr) )
		return 0;

	chblk_sz = wr->wr_tot_chunks * sizeof(struct gidx_chunk);

	DEBUG("allocated %p", wr);
	first_chunk = sizeof(struct gidx_hdr) + idxblk_sz +
			chblk_sz + strtab_sz;
	align = GIDX_ALIGN_BYTES(first_chunk);
	first_chunk = ALIGN_GIDX(first_chunk);

	INFO("%u indices, %u chunks, %u byte strtab",
		wr->wr_num_fields, wr->wr_tot_chunks, strtab_sz);
	INFO("%u byte header, %u bytes indexes, %u bytes chunks, "
		"%u bytes strtab, %u bytes align",
		sizeof(struct gidx_hdr), idxblk_sz,
		chblk_sz, strtab_sz, align);
	INFO("first chunk begins at 0x%llx", first_chunk);

	buf = fobuf_new(fd, 0);
	if ( NULL == buf )
		return 0;

	if ( !write_header(buf, wr->wr_num_fields,
				wr->wr_tot_chunks, strtab_sz) )
		goto abort;

	flayout = calloc(wr->wr_num_fields, sizeof(*flayout));
	if ( NULL == flayout )
		goto abort;
	for (i = 0; i < wr->wr_num_fields; i++)
		flayout[i] = wr->wr_field + i;
	sort(flayout, wr->wr_num_fields, sizeof(*flayout), f_cmp);

	for(s_ofs = c_idx = i = 0; i < wr->wr_num_fields; i++) {
		if ( !write_index(buf, flayout[i], &c_idx, &s_ofs) )
			goto abort;
	}

	clayout = calloc(wr->wr_tot_chunks, sizeof(*clayout));
	if ( NULL == clayout )
		goto abort;
	for (i = 0; i < wr->wr_tot_chunks; i++)
		clayout[i] = wr->wr_chunk + i;

	data_sz = layout_chunks(clayout, wr->wr_tot_chunks, first_chunk);
	if ( 0 == data_sz )
		goto abort;

	INFO("%llu bytes data, total file size expeted at 0x%.llx",
		data_sz, first_chunk + data_sz);
	for(i = 0; i < wr->wr_num_fields; i++) {
		for(j = 0; j < flayout[i]->f_num_chunks; j++) {
			if ( !write_chdr(buf, flayout[i]->f_chunk + j) )
				goto abort;
		}
	}

	for(i = 0; i < wr->wr_num_fields; i++) {
		if ( !fobuf_write(buf, flayout[i]->f_name,
					strlen(flayout[i]->f_name)) )
			goto abort;
	}

	if ( !fobuf_write(buf, alignbuf, align) )
		goto abort;

	for(i = 0; i < wr->wr_tot_chunks; i++) {
		switch(clayout[i]->c_blkid) {
		case -1:
			switch(clayout[i]->c_format) {
			case GIDX_CHUNK_PKEY_TBL:
				if ( !write_pk_tbl(buf, wr, clayout[i]) )
					goto abort_state;
				break;
			case GIDX_CHUNK_PKEY_BLOCK:
				if ( !write_pk_block(buf, wr, clayout[i]) )
					goto abort_state;
				break;
			default:
				ERR("Bad system chunk format %.2x",
					clayout[i]->c_format);
				goto abort_state;
			}
			break;
		default:
			assert(clayout[i]->c_blkid >= 0);
			assert(clayout[i]->c_blkid <= 0xffff);
			if ( !write_user_chunk(buf, clayout[i]) )
				goto abort;
			break;
		}
	}

	free(flayout);
	free(clayout);
	return fobuf_close(buf);
abort_state:
	for(i = 0; i < wr->wr_num_fields; i++) {
		free(wr->wr_field[i].f_ctx);
		wr->wr_field[i].f_ctx = NULL;
	}
abort:
	ERR("write aborted");
	fobuf_abort(buf);
	free(flayout);
	free(clayout);
	return 0;
}

_public void gidx_wr_free(gidx_wr_t wr)
{
	unsigned int i;
	DEBUG("%p", wr);
	if ( NULL == wr )
		return;

	free(wr->wr_chunk);
	for(i = 0; i < wr->wr_num_fields; i++) {
		struct _gidx_type_wr *type;
		type = wr->wr_field[i].f_type;
		(*type->t_wr_fini)(wr->wr_field[i].f_priv);
		DEBUG("free field[%u]: %s",
			i, wr->wr_field[i].f_name);
	}
	free(wr->wr_field);
	free(wr->wr_rec);
	free(wr);
}

gidx_pkey_t gidx_wr_num_pkey(gidx_wr_t wr)
{
	return wr->wr_num_records;
}

struct _gidx_type_wr *_gidx_types[GIDX_NR_TYPES] = {
	[GIDX_TYPE_STRING] 		&_gidx_str_type,
//	[GIDX_TYPE_U8]			&_gidx_u8_type,
//	[GIDX_TYPE_U16]			&_gidx_u16_type,
	[GIDX_TYPE_U32]			&_gidx_u32_type,
//	[GIDX_TYPE_U32_INTERVAL]	&_gidx_interval_type;
};
