/* TODO:
 *  - outward edges must be sorted for binary search
 *  - strtab should be compressed, it's MASSIVELY redundant
*/
#include <gidx.h>
#include <gidx-wr.h>
#include <gang.h>
#include <gidx-format.h>
#define _THIS_IS_LIBGIDX_WR 1
#include "gidx-wr-core.h"
#define _THIS_IS_GIDX_STR_WR 1
#include "gidx-str-wr.h"

#define DEBUG_INCORE 		0

struct radix {
	hgang_t			r_edges;
	struct radix_edge	*r_root_edge;
	uint32_t		r_strtab_sz;
	unsigned int		r_num_edges;
	unsigned int		r_max_fanout;
	unsigned int		r_max_cmp;
	unsigned int		r_max_path;
};

/* nodes in a radix tree do nothing but point to more edges so fuck it,
 * we just store the edges
 */
struct radix_edge {
	struct list_head	e_list;
	struct ro_vec		e_cmp;
	struct str_val		*e_terminal;
	unsigned int 		e_num_edges;
	struct list_head	e_edges;
};

static struct radix_edge *edge_new(struct radix *r, struct radix_edge *parent)
{
	struct radix_edge *e;

	e = hgang_alloc(r->r_edges);
	if ( NULL == e ) {
		ERR("%s", os_err());
		return NULL;
	}

	INIT_LIST_HEAD(&e->e_edges);
	e->e_num_edges = 0;
	e->e_terminal = NULL;

	if ( parent ) {
		list_add_tail(&e->e_list, &parent->e_edges);
		parent->e_num_edges++;
		if ( parent->e_num_edges > r->r_max_fanout )
			r->r_max_fanout = parent->e_num_edges;
	}
	r->r_num_edges++;

	return e;
}

static struct str_val *next_string(struct str_val *v, struct str_val *end)
{
	if ( v == end )
		return NULL;
	return (struct str_val *)node_next(&v->n_rbt);
}

static size_t common_prefix(struct ro_vec *a, struct ro_vec *b,
				size_t ofs)
{
	size_t ret, min;

	min = (a->v_len < b->v_len) ? a->v_len : b->v_len;
	DEBUG("%.*s %.*s ofs=%u min=%u",
		a->v_len, a->v_ptr,
		b->v_len, b->v_ptr,
		ofs, min);
	//assert(ofs < min);
	if ( ofs >= min )
		return 0;
	DEBUG("%.*s %.*s",
		min - ofs, a->v_ptr + ofs,
		min - ofs, b->v_ptr + ofs);

	for(ret = 0; ret < min - ofs; ret++)
		if ( a->v_ptr[ofs + ret] != b->v_ptr[ofs + ret] )
			break;

	return ret;
}

static int do_radix(struct radix *r, struct radix_edge *n,
			struct str_val *v, struct str_val *end, size_t ofs,
			unsigned int depth)
{
	struct str_val *tmp, *last;
	size_t len;

	if ( depth > r->r_max_path )
		r->r_max_path = depth;

	DEBUG("ofs = %u", ofs);
	while(v) {
		struct radix_edge *e;

		for(len = v->n_str.v_len - ofs, last = tmp = v;
				tmp && len < 0x100;
				last = tmp, tmp = next_string(tmp, end)) {
			size_t x;
			x = common_prefix(&v->n_str, &tmp->n_str, ofs);
			if ( 0 == x )
				break;
			len = (x < len) ? x : len;
		}

		e = edge_new(r, n);
		if ( NULL == e )
			return 0;

		e->e_cmp.v_ptr = v->n_str.v_ptr + ofs;
		e->e_cmp.v_len = len;
		r->r_strtab_sz += len;
		if ( e->e_cmp.v_len > r->r_max_cmp )
			r->r_max_cmp = e->e_cmp.v_len;
		assert(ofs + len <= v->n_str.v_len);

		if ( ofs + len == v->n_str.v_len ) {
			mesg(M_DEBUG, "radix-terminal: %.*s",
				v->n_str.v_len, v->n_str.v_ptr);
			e->e_terminal = v;
			if ( v == last ) {
				mesg(M_DEBUG, "  - no recurse");
				v = next_string(last, end);
				continue;
			}else{
				mesg(M_DEBUG, "  - recurse without first");
				v = next_string(v, end);
			}
		}else
			mesg(M_DEBUG, "  - plain recurse");

		mesg(M_DEBUG, "prefix: %.*s", ofs + len, v->n_str.v_ptr);
		for(tmp = v; tmp; tmp = next_string(tmp, end)) {
			mesg(M_DEBUG, "  shared by: %.*s",
					tmp->n_str.v_len, tmp->n_str.v_ptr);
			if ( tmp == last )
				break;
		}

		do_radix(r, e, v, last, ofs + len, depth + 1);
		v = next_string(last, end);
	}

	return 1;
}

static int construct_radix(struct str_wr_priv *strp, struct radix *r)
{
	struct str_val *v;
	struct radix_edge *n;

	v = (struct str_val *)node_first(&strp->p_vals);
	if ( NULL == v )
		return 1; /* empty radix tree */

	r->r_root_edge = n = edge_new(r, NULL);
	if ( NULL == n )
		return 0;

	return do_radix(r, n, v, NULL, 0, 1);
}

#if DEBUG_INCORE
#include <stdio.h>
static void do_dump(FILE *f, struct radix_edge *n)
{
	struct radix_edge *e;

	fprintf(f, "\t\"n_%p\" [label=\"\"];\n", n);
	if ( n->e_terminal ) {
		fprintf(f, "\t\"t_%p\" [shape=rectangle label=\"%.*s\"];\n",
			n->e_terminal,
			n->e_terminal->n_str.v_len,
			n->e_terminal->n_str.v_ptr);
		fprintf(f, "\t\"n_%p\" -> \"t_%p\";\n", n, n->e_terminal);
	}

	list_for_each_entry(e, &n->e_edges, e_list) {
		do_dump(f, e);
		fprintf(f, "\t\"n_%p\" -> \"n_%p\" [label=\"%.*s\"];\n",
			n, e, e->e_cmp.v_len, e->e_cmp.v_ptr);
	}
}

static void dump_radix(struct radix *r)
{
	FILE *f;

	f = fopen("radix.dot", "w");
	fprintf(f, "strict digraph \"Radix trie\" {\n");
	fprintf(f, "\tgraph[rankdir=LR];\n");
	fprintf(f, "\tnode[shape=ellipse, style=filled, "
			"fillcolor=transparent];\n");
	//fprintf(f, "\tedge[fontsize=6];\n");
	do_dump(f, r->r_root_edge);
	fprintf(f, "}\n");
	fclose(f);
}
#endif

int _str_radix_prep(struct str_wr_priv *strp)
{
	struct radix *r;

	r = calloc(1, sizeof(*r));
	if ( NULL == r )
		return 0;

	r->r_edges = hgang_new(sizeof(struct radix_edge), 0);
	if ( NULL == r->r_edges )
		goto out_free;

	if ( !construct_radix(strp, r) )
		goto out_free_edges;

#if DEBUG_INCORE
	dump_radix(r);
#endif

	WARN ("%u edges, %u max fanout, %u max_cmp, %u max path",
		r->r_num_edges, r->r_max_fanout,
		r->r_max_cmp, r->r_max_path);
	strp->p_fpriv = r;
	strp->p_prepped = 1;
	return 1;
out_free_edges:
	hgang_free(r->r_edges);
out_free:
	free(r);
	return 0;
}

void _str_radix_free(struct str_wr_priv *strp)
{
	struct radix *r = strp->p_fpriv;
	if ( r) {
		hgang_free(r->r_edges);
		free(r);
		strp->p_fpriv = NULL;
	}
}

uint16_t _str_radix_num_chunks(struct str_wr_priv *strp)
{
	return RADIX_NUM_CHUNKS;
}

off_t _str_radix_chunk_size(struct str_wr_priv *strp, unsigned int chunk_id)
{
	struct radix *r = strp->p_fpriv;

	switch(chunk_id) {
	case RADIX_CHUNK_EDGES:
		return ALIGN_GIDX(sizeof(struct radix_hdr)) +
				(sizeof(struct radix_dedge) * r->r_num_edges);
	case RADIX_CHUNK_STRTAB:
		return r->r_strtab_sz;
	default:
		abort();
	}
	return 0;
}

static int do_write_radix(fobuf_t buf, struct radix_edge *e,
			gidx_pkey_t *pk, gidx_pkey_t nt,
			uint32_t *idx, uint32_t *s_ofs)
{
	struct radix_edge *ee;
	struct radix_dedge d;
	uint32_t e_idx = (*idx);

	memset(&d, 0, sizeof(d));

	if ( e->e_terminal ) {
		d.re_pkey = (*pk)++;
	}else{
		d.re_pkey = nt;
	}

	/* PROOF: if there are n : n > 0xff outward edges then common prefix
	 * length must be > 1. But if common prefix is > 1 then there are at
	 * most 0xff outward edges with coomon prefix = 1. Therefore max
	 * outward edges is at most 0xff. QED
	 *
	 * FIXME: actually it's 0x100 so that's off by one. we could encode
	 * this as n - 1 and for the case outgoing edges = 0 then edges 
	 * index should be zero which is invalid because thats the root
	 * node so any reference to that would introduce a cycle
	 */
	assert(e->e_num_edges < 0x100);

	d.re_num_edges = e->e_num_edges;
	d.re_strtab_len = e->e_cmp.v_len;
	d.re_strtab_hi = ((*s_ofs) >> 24) & 0xff;
	d.re_strtab_ofs = (*s_ofs) & 0xffff;
	d.re_edges_hi = ((e_idx) > 24) & 0xff;
	d.re_edges_idx = (e_idx) & 0xffff;

	if ( !fobuf_write(buf, &d, sizeof(d)) )
		return 0;

	(*idx) += e->e_num_edges;
	(*s_ofs) += d.re_strtab_len;

	list_for_each_entry(ee, &e->e_edges, e_list)
		if ( !do_write_radix(buf, ee, pk, nt, idx, s_ofs) )
			return 0;

	return 1;
}

static int write_radix(struct str_wr_priv *strp, struct radix *r, fobuf_t buf)
{
	struct radix_hdr h;
	uint32_t idx = 1, s_ofs = 0;
	gidx_pkey_t pk = 0;

	h.h_num_vals = strp->p_num_vals;
	if ( !fobuf_write(buf, &h, sizeof(h)) )
		return 0;
	if ( !gidx_wr_align(buf, sizeof(h)) )
		return 0;

	if ( !do_write_radix(buf, r->r_root_edge, &pk,
				h.h_num_vals + 1, &idx, &s_ofs) )
		return 0;

	//INFO("idx left at %u", idx);
	assert(s_ofs == r->r_strtab_sz);
	return 1;
}

static int write_strtab(fobuf_t buf, struct radix_edge *e)
{
	struct radix_edge *ee;

	if ( !fobuf_write(buf, e->e_cmp.v_ptr, e->e_cmp.v_len) )
		return 0;

	list_for_each_entry(ee, &e->e_edges, e_list)
		if ( !write_strtab(buf, ee) )
			return 0;

	return 1;
}

int _str_radix_chunk(struct str_wr_priv *strp, unsigned int chunk_id,
			fobuf_t buf)
{
	struct radix *r = strp->p_fpriv;

	switch(chunk_id) {
	case RADIX_CHUNK_EDGES:
		return write_radix(strp, r, buf);
	case RADIX_CHUNK_STRTAB:
		return write_strtab(buf, r->r_root_edge);
	default:
		abort();
	}

	return 1;
}
